@@ -29,7 +29,8 @@ from abc import ABC, abstractmethod
 from enum import Enum
 from typing import Any, Dict, List, Optional, Tuple, Union
 
-import bcj  # type: ignore  # noqa
+import bcj
+import inflate64
 import pyppmd
 import pyzstd
 from Cryptodome.Cipher import AES
@@ -65,10 +66,6 @@ try:
     import brotli  # type: ignore  # noqa
 except ImportError:
     import brotlicffi as brotli  # type: ignore  # noqa
-try:
-    import inflate64  # noqa
-except ImportError:
-    inflate64 = None  # type: ignore  # noqa
 brotli_major = 1
 brotli_minor = 0
 
@@ -265,7 +262,25 @@ class DeflateDecompressor(ISevenZipDecompressor):
 
 class Deflate64Compressor(ISevenZipCompressor):
     def __init__(self):
-        raise RuntimeError("Deflate64 compression not implemented yet")
+        self.flushed = False
+        if inflate64 is not None:
+            self._compressor = inflate64.Deflater()
+            self._enabled = True
+        else:
+            self._enabled = False
+
+    def compress(self, data: Union[bytes, bytearray, memoryview], max_length: int = -1) -> bytes:
+        if not self._enabled:
+            raise UnsupportedCompressionMethodError(None, "inflate64 is not installed.")
+        return self._compressor.deflate(data)
+
+    def flush(self) -> bytes:
+        if not self._enabled:
+            raise UnsupportedCompressionMethodError(None, "inflate64 is not installed.")
+        if self.flushed:
+            return b""
+        self.flushed = True
+        return self._compressor.flush()
 
 
 class Deflate64Decompressor(ISevenZipDecompressor):
@@ -1142,12 +1157,6 @@ class SupportedMethods:
                 " and help Python core team implementing it."
                 " Or please use another tool to extract it.",
             )
-        if coder["method"] == COMPRESSION_METHOD.MISC_DEFLATE64:
-            raise UnsupportedCompressionMethodError(
-                coder["method"],
-                "DEFLATE64 compression is not supported by py7zr yet."
-                " Please check the progress in py7zr project home page.",
-            )
         if coder["method"] == COMPRESSION_METHOD.MISC_LZ4:
             raise UnsupportedCompressionMethodError(
                 coder["method"], "Archive is compressed by an unsupported algorythm LZ4."
