@@ -25,6 +25,8 @@ import lzma
 import platform
 import sys
 
+import psutil  # type: ignore
+
 MAGIC_7Z = binascii.unhexlify("377abcaf271c")
 FINISH_7Z = binascii.unhexlify("377abcaf271d")
 
@@ -37,15 +39,72 @@ COMMAND_HELP_STRING = """<Commands>
 """
 
 
-def get_default_blocksize():
-    if platform.python_implementation() == "PyPy" and sys.version_info >= (3, 6, 9):
-        return 1048576
-    elif sys.version_info >= (3, 7, 5):
+def is_64bit() -> bool:
+    """check if running platform is 64bit python."""
+    return sys.maxsize > 1e32
+
+
+def is_pypy369later() -> bool:
+    """check if running platform is PyPY and python 3.6.9 and later"""
+    return platform.python_implementation() == "PyPy" and sys.version_info >= (3, 6, 9)
+
+
+def get_default_blocksize() -> int:
+    """
+    Return a safe buffer size for pass to decompress/compress modules.
+    We check several conditions;
+
+    1. 32bit python
+      It sometimes fails with memory_error
+      when decompress/compress large file (hundreds of MiB) on 32bit python.
+      We reduce buffer size smaller to avoid memory_error.
+      @see issue https://github.com/miurahr/py7zr/issues/370
+
+    2. CPython
+      CPython implementation of 3.7.5 fixed a lzma module bug
+      that is not respect max_length parameter.
+      When buffer size is larger than default of the module,
+      max_length can be a value to lead the bug, so we set it
+      lzma module default = 32768 bytes.
+      @see BPO-21872: LZMA library sometimes fails to decompress a file
+           https://bugs.python.org/issue21872
+      @see https://github.com/miurahr/py7zr/issues/272
+
+    3. PyPy
+      PyPy 7.2 (python 3.6.9) fixed a lzma module's bug as same as
+      CPython above.
+      We set buffer size is as default size of the module to avoid the bugs.
+      @see PyPy3-3090: lzma.LZMADecomporessor.decompress does not respect max_length
+           https://foss.heptapod.net/pypy/pypy/-/issues/3090
+      @see https://github.com/miurahr/py7zr/pull/114
+
+    :return: recommended buffer size as int.
+    """
+    if is_64bit() and (is_pypy369later() or sys.version_info >= (3, 7, 5)):
         return 1048576
     else:
         return 32768
 
 
+def get_memory_limit():
+    """
+    Get memory limit for allocating decompression chunk buffer.
+    :return: allowed chunk size in bytes.
+    """
+    default_limit = int(128e6)
+    if sys.platform.startswith("win"):
+        return default_limit
+    else:
+        import resource
+
+        soft, _ = resource.getrlimit(resource.RLIMIT_AS)
+        if soft == -1:
+            avmem = psutil.virtual_memory().available
+            return min(default_limit, (avmem - int(256e6)) >> 2)
+        else:
+            return min(default_limit, (soft - int(256e6)) >> 2)
+
+
 # Exposed constants
 FILTER_LZMA = lzma.FILTER_LZMA1
 FILTER_LZMA2 = lzma.FILTER_LZMA2
@@ -74,6 +133,7 @@ FILTER_COPY = 0x33
 FILTER_ZSTD = 0x35
 FILTER_PPMD = 0x36
 FILTER_BROTLI = 0x37
+FILTER_DEFLATE64 = 0x38
 
 
 class Constant:
